import  java.util.HashMap;
import  java.util.Map;
import  java.util.Properties;
import  java.util.Random;

import  javax.jms.Connection;
import  javax.jms.ConnectionFactory;
import  javax.jms.DeliveryMode;
import  javax.jms.Destination;
import  javax.jms.JMSException;
import  javax.jms.MapMessage;
import  javax.jms.MessageConsumer;
import  javax.jms.MessageProducer;
import  javax.jms.Session;

import  javax.naming.Context;
import  javax.naming.InitialContext;
import  javax.naming.NamingException;

import  org.apache.log4j.BasicConfigurator;
import  org.apache.log4j.ConsoleAppender;
import  org.apache.log4j.Level;
import  org.apache.log4j.Logger;
import  org.apache.log4j.PatternLayout;

/**
 * JMS load test class
 * @author rschuetz
 */

public class LoadTest
{
 // === Static ================================================================

 private static final Logger               LOGGER          = Logger.getLogger(LoadTest.class);
 private static final Map<Integer, byte[]> MESSAGES        = new HashMap<Integer, byte[]>();
 private static final Random               RANDOM          = new Random();
 private static final long                 RECEIVE_TIMEOUT = 30000L;
 
 // === Attributes ============================================================

 private final String                jndiFactoryInitial,
                                     jndiProviderURL,
                                     jndiConnectionFactoryAlias,
                                     jndiQueueAlias;

 private final long                  totalRunTime;

 private final int                   msgCount,
                                     msgCountVariance,
                                     stageLoops,
                                     stages,
                                     minMsgSize,
                                     maxMsgSize,
                                     commitInterval,
                                     reconnectInterval; 
 
 private       Connection            producerConnection       = null,
                                     consumerConnection       = null;
 private       Session               producerSession          = null,
                                     consumerSession          = null;
 private       MessageProducer       messageProducer          = null;
 private       MessageConsumer       messageConsumer          = null;
 
 private       long                  currentQueueSize         = 0L;
 
 private       long                  totalNbrMessagesSent     = 0L,
                                     totalNbrMessagesReceived = 0L;
  
 // === Constructor ===========================================================
 
 public LoadTest(String[] args)
 {
  int argIdx;
  
  // Initialize logger
  
  BasicConfigurator.resetConfiguration();

  Logger.getRootLogger().addAppender(new ConsoleAppender(new PatternLayout("%-5p [%d{ISO8601}] %m%n")));
  Logger.getRootLogger().setLevel(Level.INFO);
  
  // Get parameters
  
  if(args.length == 13)
  {
   argIdx                     = 0;
   
   jndiFactoryInitial         = args[argIdx ++];
   jndiProviderURL            = args[argIdx ++];
   jndiConnectionFactoryAlias = args[argIdx ++];
   jndiQueueAlias             = args[argIdx ++];
   
   totalRunTime               = Long   .parseLong(args[argIdx ++]);
   msgCount                   = Integer.parseInt (args[argIdx ++]);
   msgCountVariance           = Integer.parseInt (args[argIdx ++]);
   stages                     = Integer.parseInt (args[argIdx ++]);   
   stageLoops                 = Integer.parseInt (args[argIdx ++]);
   minMsgSize                 = Integer.parseInt (args[argIdx ++]);
   maxMsgSize                 = Integer.parseInt (args[argIdx ++]);
   commitInterval             = Integer.parseInt (args[argIdx ++]);
   reconnectInterval          = Integer.parseInt (args[argIdx ++]);
  }
  else
  {
   throw new IllegalArgumentException("Usage: " + LoadTest.class.getSimpleName() + " <jndi factory initial> <jndi provider URL> <jndi connection factory alias> <jndi queue alias> <the connection to use> <total runtime in ms> <max. message count in a queue (+/- variance)> <message count variance> <# of stages> <# of loops adding/removing # of variance messages on queue per stage> <minimum message size in bytes> <maximum message size in bytes> <commit interval in # of messages processed> <reconnect interval in # of messages processed>");
  }
 }
 
 // === Public ================================================================

 public static void main(String[] args)
 throws Exception
 {
  try
  {
   (new LoadTest(args)).runLoadTest();
  }
  catch(Exception e)
  {
   LOGGER.error("ERROR", e);
  }
 }
 public void runLoadTest()
 throws Exception
 {
  int  stage;
  long startTime, stageLimit, nbrMessagesToSend, nbrMessagesToReceive;
  
  // Initialize
  
  startTime        = System.currentTimeMillis();
  currentQueueSize = 0L;
  
  // Initial connect
  
  connect();
  
  // Loop until specified runtime elapsed
  
  while(System.currentTimeMillis() < startTime + totalRunTime)
  {
   // Fill queue in stages
   
   for(stage = 0; stage <= stages; stage ++)
   {
    // Fill queue to stage limit
    
    stageLimit        = (msgCount * stage) / stages;
    nbrMessagesToSend = stageLimit - currentQueueSize;
    
    LOGGER.info("FILL: Filling queue to stage " + stage + " limit " + stageLimit);  
    
    sendMessages(nbrMessagesToSend);
    
    // Apply variance
    
    for(int stageLoop = 0; stageLoop < stageLoops; stageLoop ++)
    {
     if(stageLimit - msgCountVariance < 0)
     {
      LOGGER.info("FILL/VAR: Stage: " + stage + "/" + stageLoop + "; sending.");  
      sendMessages(msgCountVariance);
      
      LOGGER.info("FILL/VAR: Stage: " + stage + "/" + stageLoop + "; receiving.");  
      receiveMessages(msgCountVariance);
     }
     else
     {
      LOGGER.info("FILL/VAR: Stage: " + stage + "/" + stageLoop + "; sending [1].");  
      sendMessages(msgCountVariance);
     
      LOGGER.info("FILL/VAR: Stage: " + stage + "/" + stageLoop + "; receiving.");  
      receiveMessages(msgCountVariance * 2);
     
      LOGGER.info("FILL/VAR: Stage: " + stage + "/" + stageLoop + "; sending [2].");  
      sendMessages(msgCountVariance);
     }
    }
   }
   
   // Print totals
   
   LOGGER.info("FILL: Totals: # msgs sent: " + totalNbrMessagesSent + " # msgs received: " + totalNbrMessagesReceived);   
   
   // Empty queue in stages
   
   for(stage = stages; stage >= 0; stage --)
   {
    // Empty queue to stage limit
    
    stageLimit           = (msgCount * stage) / stages;
    nbrMessagesToReceive = currentQueueSize - stageLimit;
    
    LOGGER.info("EMPY: Emptying queue to stage " + stage + " limit " + stageLimit);  
    
    receiveMessages(nbrMessagesToReceive);
    
    // Apply variance
    
    for(int stageLoop = 0; stageLoop < stageLoops; stage ++)
    {
     if(stageLimit - msgCountVariance < 0)
     {
      LOGGER.info("EMPY/VAR: Stage: " + stage + "/" + stageLoop + "; sending.");  
      sendMessages(msgCountVariance);
      
      LOGGER.info("EMPY/VAR: Stage: " + stage + "/" + stageLoop + "; receiving.");  
      receiveMessages(msgCountVariance);
     }
     else
     {
      LOGGER.info("EMPY/VAR: Stage: " + stage + "/" + stageLoop + "; sending [1].");  
      sendMessages(msgCountVariance);
     
      LOGGER.info("EMPY/VAR: Stage: " + stage + "/" + stageLoop + "; receiving.");  
      receiveMessages(msgCountVariance * 2);
     
      LOGGER.info("EMPY/VAR: Stage: " + stage + "/" + stageLoop + "; sending [2].");  
      sendMessages(msgCountVariance);
     }
    }
   }
   
   // Print totals
   
   LOGGER.info("EMPY: Totals: # msgs sent: " + totalNbrMessagesSent + " # msgs received: " + totalNbrMessagesReceived);
  }
 }
 
 private void sendMessages(final long nbrMessages) 
 throws JMSException, 
        NamingException
 {
  int        msgSize;
  long       nbrMessagesSent;
  
  byte[]     payload;
  MapMessage mapMessage;
  
  // Log
  
  LOGGER.info("Sending " + nbrMessages + " messages, current queue size: " + currentQueueSize);
  
  // Process messages
  
  nbrMessagesSent = 0L;
  
  for(long msgCounter = 0; msgCounter < nbrMessages; msgCounter ++)
  {
   // Disconnect/Reconnect
   
   if((totalNbrMessagesSent + totalNbrMessagesReceived) % reconnectInterval == 0) reconnect();
     
   // Send
   // Get/Create payload
   
   msgSize = minMsgSize + RANDOM.nextInt(maxMsgSize - minMsgSize + 1);
   
   if((payload = MESSAGES.get(msgSize)) == null)
   {
    RANDOM  .nextBytes(payload = new byte[msgSize]);
    MESSAGES.put(msgSize, payload);
   }
   
   // Create message

   mapMessage = producerSession.createMapMessage();
   
   mapMessage.setStringProperty("messageid", this.getClass().getSimpleName() + "#" + totalNbrMessagesSent);
   mapMessage.setLongProperty  ("timestamp", System.currentTimeMillis());
   mapMessage.setStringProperty("route",     "gprs");

   mapMessage.setString        ("address", "LOADTEST");
   mapMessage.setInt           ("type",    10);
   mapMessage.setBytes         ("data",    payload);

   // Send message

   messageProducer.send(mapMessage);
   
   nbrMessagesSent      ++;
   totalNbrMessagesSent ++;
   
   // Commit
   
   if(totalNbrMessagesSent % commitInterval == 0) 
   {
    LOGGER.info("Committing producer session [intermediate].");
    producerSession.commit();
   }
  }
  
  // Commit
  
  LOGGER.info("Committing producer session [end send loop].");
  producerSession.commit();
  
  // Update queue size
  
  currentQueueSize += nbrMessagesSent;
  
  // Check
  
  if(nbrMessagesSent != nbrMessages) throw new IllegalStateException("Send: expected: " + nbrMessages + " actual: " + nbrMessagesSent);
  
  // Log
  
  LOGGER.info("Done; " + nbrMessagesSent + " messages sent, current queue size: " + currentQueueSize);
 }
 
 private void receiveMessages(final long nbrMessages) 
 throws JMSException, 
        NamingException
 {
  long nbrMessagesReceived;
  
  // Log
  
  LOGGER.info("Receiving " + nbrMessages + " messages, current queue size: " + currentQueueSize);  
  
  // Process messages
  
  nbrMessagesReceived = 0L;
  
  for(long msgCounter = 0; msgCounter < nbrMessages; msgCounter ++)
  {
   // Disconnect/Reconnect
   
   if((totalNbrMessagesSent + totalNbrMessagesReceived) % reconnectInterval == 0) reconnect();
     
   // Receive
   
   if(messageConsumer.receive(RECEIVE_TIMEOUT) != null) 
   { 
    nbrMessagesReceived      ++;
    totalNbrMessagesReceived ++;  
   }
   else break;
   
   // Commit
   
   if(totalNbrMessagesReceived % commitInterval == 0) 
   {
    LOGGER.info("Committing consumer session [intermediate].");  
    consumerSession.commit();
   }
  }
  
  // Commit
  
  LOGGER.info("Committing consumer session [end recv loop].");  
  consumerSession.commit();
  
  // Update queue size
  
  currentQueueSize -= nbrMessagesReceived;
  
  // Check
  
  if(nbrMessagesReceived != nbrMessages) throw new IllegalStateException("Receive: expected: " + nbrMessages + " actual: " + nbrMessagesReceived);
  
  // Log
  
  LOGGER.info("Done. " + nbrMessagesReceived + " messages received, current queue size: " + currentQueueSize);
 }
 
 // === Private ===============================================================

 private void reconnect() 
 throws JMSException, 
        NamingException
 {
  disconnect();
  connect();
 }
 
 private void connect() 
 throws JMSException, 
        NamingException
 {
  Properties        jndiProperties;
  InitialContext    jndiContext;
  ConnectionFactory connectionFactory;
  Destination       destination;
   
  // Log
  
  LOGGER.info("Connecting.");  
  
  // Initialize JNDI

  jndiProperties = new Properties();
  jndiProperties.setProperty(Context.INITIAL_CONTEXT_FACTORY, jndiFactoryInitial);
  jndiProperties.setProperty(Context.PROVIDER_URL,            jndiProviderURL);

  jndiContext = new InitialContext(jndiProperties);

  // Lookup connection factory and queue

  connectionFactory = ((ConnectionFactory)(jndiContext.lookup(jndiConnectionFactoryAlias)));
  destination       = ((Destination      )(jndiContext.lookup(jndiQueueAlias)));

  // Create producer connection, session and message producer

  producerConnection = connectionFactory.createConnection();
  producerSession    = producerConnection.createSession(true, 0);
  
  messageProducer = producerSession.createProducer(destination);
  messageProducer.setDeliveryMode(DeliveryMode.PERSISTENT); 
  
  producerConnection.start();  
  
  // Create consumer connection, session and message consumer

  consumerConnection = connectionFactory.createConnection();
  consumerSession    = consumerConnection.createSession(true, 0);
  
  messageConsumer = consumerSession.createConsumer(destination);
  
  consumerConnection.start();
  
  // Log
  
  LOGGER.info("Connected.");    
 }
 
 private void disconnect() 
 throws JMSException
 {
  // Log
  
  LOGGER.info("Disconnecting.");    
  
  // Close assets
  
  if(messageProducer    != null) {                           messageProducer   .close(); messageProducer    = null; }  
  if(messageConsumer    != null) {                           messageConsumer   .close(); messageConsumer    = null; }
  if(producerSession    != null) { producerSession.commit(); producerSession   .close(); producerSession    = null; }  
  if(consumerSession    != null) { consumerSession.commit(); consumerSession   .close(); consumerSession    = null; }
  if(producerConnection != null) {                           producerConnection.close(); producerConnection = null; }
  if(consumerConnection != null) {                           consumerConnection.close(); consumerConnection = null; }  
  
  // Log
  
  LOGGER.info("Disconnected.");      
 }
 
 // === Inner classes =========================================================
}
